"""Reconnecting websocket client"""

import asyncio
import logging
from typing import Callable, Coroutine

from websockets.asyncio.client import ClientConnection, connect
from websockets.exceptions import ConnectionClosed
from websockets.protocol import State

from .task import TaskList, task_manager

_LOGGER = logging.getLogger(__name__)


class ReconnectingWebsocketClient:
    """Reconnecting websocket client"""

    def __init__(
        self,
        host: str,
        port: int,
        on_message: Callable[[str], Coroutine[None, str, None]],
        on_state_change: Callable[[str], Coroutine[None, State, None]] | None = None,
        on_connect: Callable[[], Coroutine[None, None, None]] | None = None,
        connect_timeout=15,
    ):
        self._host = host
        self._port = port
        self._connect_timeout = connect_timeout
        self._outgoing_queue: asyncio.Queue[str] = asyncio.Queue()
        self._on_message = on_message
        self._on_state_change = on_state_change
        self._on_connect = on_connect
        self._tasks = TaskList()
        self._ws = None

    @property
    def state(self) -> State | None:
        """State of connection"""
        if self._ws:
            return self._ws.protocol.state
        return None

    async def __do_on_state_change(self):
        try:
            if callable(self._on_state_change):
                if isinstance(result := self._on_state_change(self.state), Coroutine):
                    await result
        except BaseException:
            _LOGGER.exception("Error calling on_state_change")

    async def __do_on_message(self, message: str):
        try:
            if callable(self._on_message):
                if isinstance(result := self._on_message(message), Coroutine):
                    await result
        except BaseException:
            _LOGGER.exception("Error calling on_message")

    async def __do_on_connect(self):
        try:
            if callable(self._on_connect):
                if isinstance(result := self._on_connect(), Coroutine):
                    await result
        except BaseException:
            _LOGGER.exception("Error calling on_connect")

    async def send(self, message: str) -> None:
        """Add message to send queue"""
        await self._outgoing_queue.put(message)

    async def __runner(self):
        """Send and receive messages on websocket"""
        try:
            uri = f"ws://{self._host}:{self._port}"
            _LOGGER.info("Connecting to %s", uri)
            async for websocket in connect(
                uri, open_timeout=self._connect_timeout, ping_interval=None
            ):
                try:
                    self._ws = websocket
                    _LOGGER.info("Connection established to %s", uri)
                    await self.__do_on_connect()
                    await self.__do_on_state_change()
                    async with task_manager(cancel_on_exit=True) as ws_tasks:
                        pong_task = asyncio.create_task(
                            self.__keepalive(websocket), name="pong"
                        )
                        consumer_task = asyncio.create_task(
                            self.__consumer(websocket), name="consumer"
                        )
                        producer_task = asyncio.create_task(
                            self.__producer(websocket), name="producer"
                        )
                        ws_tasks.add(consumer_task, producer_task, pong_task)
                        await ws_tasks.wait(return_when=asyncio.FIRST_COMPLETED)
                except ConnectionClosed as e:  # pylint: disable=W0718
                    _LOGGER.warning(
                        "Connection to %s closed by remote host: %s, will retry",
                        uri,
                        e.reason,
                    )
                    continue
                except Exception as e:
                    _LOGGER.error("Error occurred in websocket: %s", e)
                    raise
                finally:
                    await self.__do_on_state_change()
        except asyncio.CancelledError:
            _LOGGER.debug("Shutting down connection to %s", uri)
            raise

    async def __consumer(self, ws: ClientConnection):
        """Enqueue messages received on websocket"""
        try:
            async for message in ws:
                if isinstance(message, str):
                    _LOGGER.debug("Message received %s", message)
                    await self.__do_on_message(message)
        except asyncio.CancelledError:
            _LOGGER.debug("Consumer was cancelled")
            raise

    async def __producer(self, ws: ClientConnection):
        """Send queued messages on websocket"""
        try:
            while True:
                message = await self._outgoing_queue.get()
                _LOGGER.debug("Sending message %s", message)
                await ws.send(message)
                self._outgoing_queue.task_done()
        except asyncio.CancelledError:
            _LOGGER.debug("Producer cancelled")
            raise

    async def __keepalive(self, websocket, pong_interval=float(30)):
        while True:
            await asyncio.sleep(pong_interval)
            try:
                _LOGGER.debug("Sending keepalive PONG")
                await websocket.send("PONG\r")
            except ConnectionClosed:
                _LOGGER.warning("Connection closed during keepalive, stopping")
                raise

    async def connect(self):
        """Connect to server"""
        if not self._tasks:
            runner_task = asyncio.create_task(self.__runner(), name="runner")
            self._tasks.add(runner_task)
        else:
            _LOGGER.warning("Already connected to %s:%s", self._host, self._port)

    def close(self):
        """Close connection and perform clean up"""
        self._tasks.cancel()

    def __enter__(self):
        self.connect()

    def __exit__(self, *args, **kwargs):
        self.close()



import asyncio
import logging
from contextlib import asynccontextmanager

_LOGGER = logging.getLogger(__name__)


@asynccontextmanager
async def task_manager(*args, cancel_on_exit=False, **kwargs):
    """Context manager for tasks

    :param cancel_on_exit: cancel tasks on exit context, defaults to False
    :type cancel_on_exit: bool, optional
    :yield: the task list
    :rtype: TaskList
    """
    task_list = TaskList(*args, **kwargs)
    try:
        yield task_list
    finally:
        if cancel_on_exit:
            task_list.cancel()


class TaskList:
    """Helper class to keep references and work with Tasks"""

    def __bool__(self):
        """Check if there are tasks in the list"""
        return bool(self.tasks)

    def __len__(self):
        return len(self.tasks)

    def __init__(self) -> None:
        self.tasks: list[asyncio.Task] = list()

    def add(self, *tasks: asyncio.Task, remove_when_done=True) -> None:
        """Add reference to tasks

        :param tasks: tasks to add
        :type tasks: `~asyncio.Task`
        :param remove_when_done: remove reference when task completes, defaults to True
        :type remove_when_done: bool, optional
        """
        for task in tasks:
            if remove_when_done:
                task.add_done_callback(self.remove)
                self.tasks.append(task)

    def remove(self, *tasks: asyncio.Task) -> None:
        """Remove references to tasks
        :param tasks: tasks to remove
        :type tasks: asyncio.Task
        """
        for task in tasks:
            try:
                self.tasks.remove(task)
            except ValueError:
                _LOGGER.exception("Failed to remove task %s", task)

    async def wait(self, *args, **kwargs) -> None:
        """Wait for tasks. Wrapper around :func:`~asyncio.wait`"""
        try:
            await asyncio.wait(self.tasks, *args, **kwargs)
        except ValueError:
            _LOGGER.exception("Failed to wait for tasks")
            pass

    def cancel(self):
        """Cancel all tasks"""
        for task in self.tasks:
            task.cancel()



"""HRV System client"""

import asyncio
import logging
from typing import Callable, Coroutine

from websockets.protocol import State

from .const import DataKey, MessageContext
from .data import Message, ParseError, UnsupportedMessageType
from .helpers.error_cache import ErrorCache
from .helpers.websocket import ReconnectingWebsocketClient

_LOGGER: logging.Logger = logging.getLogger(__name__)


class Client:
    """Client to manage communication with HRV"""

    def __init__(
        self, ip: str, port: int = 3001, update_interval=30, connect_timeout=15
    ):
        """Initiate client

        :param ip: ip address of the unit
        :type ip: str
        :param port: port
        :type port: int
        :param update_interval: update interval for calling handler, defaults to 30
        :type update_interval: int, optional
        :param connect_timeout: timeout when establishing connection, defaults to 15
        :type connect_timeout: int, optional
        """
        self._update_interval = update_interval
        self._ip = ip
        self._port = port
        self._data: dict[DataKey, str] = {}
        self._error_cache = ErrorCache()
        self._on_data_handlers: set[
            Callable[[dict[DataKey, str]], None | Coroutine]
        ] = set()
        self._on_state_change_handlers: set[Callable[[State], None | Coroutine]] = set()
        self._connect_timeout = connect_timeout
        self._tasks = [asyncio.create_task(self._do_call_data_handlers())]
        self._websocket = ReconnectingWebsocketClient(
            host=self._ip,
            port=self._port,
            connect_timeout=self._connect_timeout,
            on_message=self._on_message,
            on_connect=self._send_start_message,
            on_state_change=self._on_state_change,
        )

    @property
    def state(self):
        """State of the underlying websocket connection"""
        return self._websocket.state

    @property
    def data(self):
        """Get data from system if connection is alive"""
        if self.state == State.OPEN:
            return self._data

        return dict()

    async def connect(self):
        """Connect to HRV and begin receiving"""
        await self._websocket.connect()

    async def _send_start_message(self) -> None:
        """Send start message to server to begin receiving data"""
        message = Message(DataKey.NONE, "")
        await self._websocket.send(message.encode())

    async def _do_call_data_handlers(self):
        """Call message handlers with data at update_interval"""
        while True:
            await asyncio.sleep(self._update_interval)
            await self._call_data_handlers()

    async def _call_data_handlers(self):
        """Call handlers with data asynchronously"""
        for handler in self._on_data_handlers:
            try:
                if isinstance(result := handler(self.data), Coroutine):
                    await result
            except BaseException:
                _LOGGER.exception("Failed to call handler %s", handler)

    async def _call_state_change_handlers(self, state):
        """Call state change handlers asynchronously"""
        for handler in self._on_state_change_handlers:
            try:
                if isinstance(result := handler(state), Coroutine):
                    await result
            except BaseException:
                _LOGGER.error("Failed to call handler %s", handler)

    def close(self):
        """Disconnect from system"""
        if self._websocket:
            self._websocket.close()
        for task in self._tasks:
            task.cancel()

    async def _on_state_change(self, state):
        await self._call_state_change_handlers(state)

    async def _on_message(self, msg: str) -> None:
        """Update data"""
        try:
            message = Message.decode(msg)
            if error := self._error_cache.handle(message):
                self._data[DataKey.ERROR_MESSAGE] = str(error)
            else:
                self._data[message.key] = message.payload
                if message.message_context == MessageContext.ACK_OK:
                    await self._call_data_handlers()
        except ParseError as e:
            _LOGGER.error(e, exc_info=True)
        except UnsupportedMessageType as e:
            _LOGGER.debug("Unsupported message type: %s", e, exc_info=True)

    def add_state_change_handler(self, handler: Callable[[State], None | Coroutine]):
        """Add state change handler to be called when client state changes

        :param handler: handler to be added
        :type handler: Callable[[str], None | Coroutine]
        """
        self._on_state_change_handlers.add(handler)

    def remove_state_change_handler(self, handler: Callable[[State], None | Coroutine]):
        """Remove state change handler

        :param handler: handler to be removed
        :type handler: Callable[[str], None | Coroutine]
        """
        self._on_state_change_handlers.remove(handler)

    def add_data_handler(
        self, handler: Callable[[dict[DataKey, str]], None | Coroutine]
    ):
        """Add data handler to be called at update interval

        :param handler: handler function. Must be safe to call from event loop
        :type handler: Callable[[dict[DataKey, str]], None | Coroutine]
        """

        self._on_data_handlers.add(handler)

    def remove_data_handler(
        self, handler: Callable[[dict[DataKey, str]], None | Coroutine]
    ):
        """Remove data handler

        :param handler: handler to remove
        :type handler: Callable[[dict[DataKey, str]], None | Coroutine]
        """
        self._on_data_handlers.remove(handler)

    async def send_command(self, key: DataKey, payload: str | int):
        """Send command to HRV unit

        :param key: message type key
        :type key: MessageType
        :param payload: payload
        :type value: str | int
        """
        message = Message(key, str(payload))

        async def ack_command():
            """Should probably ack command here, just sleep for now"""
            await asyncio.sleep(0.5)

        await self._websocket.send(message.encode())
        await asyncio.gather(ack_command())

    async def __aenter__(self, *args, **kwargs):
        await self.connect()
        return self

    async def __aexit__(self, *args, **kwargs):
        self.close()
        return None



import logging

import pytest

from pysaleryd.const import DataKey, MessageContext, MessageSeparator
from pysaleryd.data import Message, SystemProperty, UnsupportedMessageType

__author__ = "Björn Dalfors"
__copyright__ = "Björn Dalfors"
__license__ = "MIT"


_LOGGER = logging.getLogger(__name__)


def test_parse_int_from_list_str():
    """Test parsing int list"""
    m = Message.decode("#MF: 1+ 0+ 2+30\r")
    assert m.key == DataKey.MODE_FAN
    assert isinstance(m.payload, str)
    assert m.message_context == MessageContext.NONE


def test_parse_ack():
    m = Message.decode("#$MF: 1\r")
    assert m.key == DataKey.MODE_FAN
    assert isinstance(m.payload, str)
    assert m.message_context == MessageContext.ACK_OK


def test_parse_ack_error():
    message = Message.decode("#!MF:\r")
    assert message.key == "MF"
    assert isinstance(message.payload, str)
    assert message.message_context == MessageContext.ACK_ERROR


def test_parse_int_from_str():
    """Test parsing int"""
    message = Message.decode("#*SC:0\r")
    assert message.key == DataKey.CONTROL_SYSTEM_VERSION
    assert isinstance(message.payload, str)
    assert message.payload == "0"
    assert message.message_context == MessageContext.NONE


def test_parse_str_from_str():
    """Test parsing str"""
    om = Message(DataKey.MODEL_NAME, "TestModel")
    message = Message.decode(om.encode())
    assert message.key == DataKey.MODEL_NAME
    assert isinstance(message.payload, str)
    assert message.payload == "TestModel"
    assert message.message_context == MessageContext.NONE

    message = Message.decode(
        f"{MessageSeparator.MESSAGE_START}{DataKey.CONTROL_SYSTEM_VERSION}{MessageSeparator.PAYLOAD_START}    1.x.1{MessageSeparator.MESSAGE_END}"  # noqa: E501
    )
    assert message.key == DataKey.CONTROL_SYSTEM_VERSION
    assert message.payload == "1.x.1"
    assert message.message_context == MessageContext.NONE


def test_parse_error():
    """Test parse error"""
    with pytest.raises(UnsupportedMessageType):
        Message.decode("wer")


def test_parse_system_property():
    """Test parse SystemProperty"""
    m = Message(DataKey.INSTALLER_PASSWORD, "1")
    parsed = SystemProperty.from_message(m)
    assert parsed.key == DataKey.INSTALLER_PASSWORD
    assert parsed.value == 1

    m.payload = "1+ 2+ 3+0"
    parsed = SystemProperty.from_str(m.key, m.payload)
    assert parsed.value == 1
    assert parsed.min_value == 2
    assert parsed.max_value == 3

    m.payload = "test"
    parsed = SystemProperty.from_str(m.key, m.payload)
    assert parsed.value == "test"


def test_parse_system_property_with_extra():
    """Test parse SystemProperty with extra value"""
    m = Message(DataKey.INSTALLER_PASSWORD, "1+ 2+ 3+ 4+")
    parsed = SystemProperty.from_str(m.key, m.payload)
    assert parsed.key == DataKey.INSTALLER_PASSWORD
    assert parsed.value == 1
    assert parsed.min_value == 2
    assert parsed.max_value == 3
    assert parsed.extra == 4



"""Utils"""

from __future__ import annotations

import logging

from .const import DataKey, MessageContext, MessageSeparator

_LOGGER: logging.Logger = logging.getLogger(__package__)


class ParseError(BaseException):
    """Parse error. Raised when message parsing fails"""


class UnsupportedMessageType(BaseException):
    """Unsupported message type. Raised when message type is not supported"""


class BaseMessage:
    """Base message class"""

    def __init__(
        self,
        key: str | DataKey,
        payload: str,
        message_context: MessageContext = MessageContext.NONE,
    ):
        self.key = DataKey(key)
        self.payload = payload
        self.message_context = message_context


class Message(BaseMessage):
    """Message from HRV system"""

    @classmethod
    def decode(cls, msg: str) -> Message:
        """Decode message from string"""
        try:
            message_context = MessageContext.from_str(msg)
            msg = msg[1::]  # remove first character (message start)

            [key, payload] = [
                v.strip() for v in msg.split(MessageSeparator.PAYLOAD_START, 1)
            ]

            if message_context in [MessageContext.ACK_OK, MessageContext.ACK_ERROR]:
                key = DataKey(key[1::])  # remove first character (ACK type)

            return cls(key, payload, message_context)
        except ValueError as exc:
            raise UnsupportedMessageType() from exc
        except Exception as exc:
            raise ParseError(f"Failed to parse message {msg}") from exc

    def encode(self) -> str:
        """Encode message to string"""
        ps = MessageSeparator
        context = self.message_context._encode()
        return f"{ps.MESSAGE_START}{context}{self.key}{ps.PAYLOAD_START}{self.payload}{ps.MESSAGE_END}"  # noqa: E501


class SystemProperty:
    """HRV System property with value, min, max"""

    def __init__(
        self,
        key: DataKey,
        value: int | str | float | None = None,
        min_value: int | str | float | None = None,
        max_value: int | str | float | None = None,
        extra: int | str | float | None = None,
    ):
        self.key = key
        self.value = value
        self.min_value = min_value
        self.max_value = max_value
        self.extra = extra

    @classmethod
    def from_message(cls, message: Message) -> SystemProperty:
        return cls.from_str(message.key, message.payload)

    @classmethod
    def from_str(cls, key: DataKey, payload: str) -> SystemProperty:
        """Create instance from from string"""

        def maybe_cast(x: str) -> int | float | str | None:
            """Cast value if it is a numeric like string"""
            if x is None:
                return x
            if x.isdigit():
                return int(x)
            if x.replace(".", "", 1).isdigit():
                return float(x)
            return x

        [*positions] = [maybe_cast(v.strip()) for v in payload.split("+")]

        value = positions[0] if len(positions) > 0 else None
        min_value = positions[1] if len(positions) > 1 else None
        max_value = positions[2] if len(positions) > 2 else None
        extra = positions[3] if len(positions) > 3 else None
        return cls(key, value, min_value, max_value, extra)


class ErrorSystemProperty:
    """HRV System error property"""

    def __init__(self, key: DataKey, value: list[str] | None = None):
        self.key = key
        self.value = value



"""Constants"""

from __future__ import annotations

from enum import StrEnum


class MessageSeparator(StrEnum):
    """Message payload separator"""

    ACK_OK = "$"
    ACK_ERROR = "!"
    MESSAGE_START = "#"
    PAYLOAD_START = ":"
    MESSAGE_END = "\r"


class MessageContext(StrEnum):
    """Message context"""

    ACK_OK = "ACK_OK"
    ACK_ERROR = "ACK_ERROR"
    NONE = "NONE"

    @classmethod
    def from_str(cls, message: str) -> MessageContext:
        """Decode message context from string"""
        pos1 = message[0]
        pos2 = message[1]
        if pos1 == MessageSeparator.MESSAGE_START:
            if pos2 == MessageSeparator.ACK_OK:
                return MessageContext.ACK_OK
            elif pos2 == MessageSeparator.ACK_ERROR:
                return MessageContext.ACK_ERROR
            else:
                return MessageContext.NONE
        else:
            raise ValueError(f"Invalid message context: {message}")

    def _encode(self) -> str:
        """Encode message context to string"""
        if self == MessageContext.ACK_OK:
            return MessageSeparator.ACK_OK
        elif self == MessageContext.ACK_ERROR:
            return MessageSeparator.ACK_ERROR
        elif self == MessageContext.NONE:
            return ""
        else:
            raise ValueError(f"Invalid message context: {self}")


class DataKey(StrEnum):
    """Data key enumerator"""

    # Scalars
    AIR_TEMPERATURE_AT_HEATER = "*TK"
    AIR_TEMPERATURE_SUPPLY = "*TC"
    CONTROL_SYSTEM_VERSION = "*SC"
    ERROR_FRAME_END = "*EZ"
    ERROR_MESSAGE = "*EB"
    ERROR_FRAME_START = "*EA"
    FAN_SPEED_EXHAUST = "*DB"
    FAN_SPEED_SUPPLY = "*DA"
    FILTER_MONTHS_LEFT = "*FL"
    HEAT_EXCHANGER_ROTOR_PERCENT = "*XA"
    HEAT_EXCHANGER_ROTOR_RPM = "*XB"
    HEATER_POWER_PERCENT = "*MJ"
    MINUTES_LEFT_BOOST_MODE = "*FI"
    MINUTES_LEFT_FIREPLACE_MODE = "*ME"
    MODEL_NAME = "*SB"
    PRODUCT_NUMBER = "*SA"
    INSTALLER_EMAIL = "IB"
    INSTALLER_NAME = "IA"
    INSTALLER_PASSWORD = "IP"
    INSTALLER_PHONE = "IC"
    INSTALLER_WEBSITE = "IE"

    # Vectors x+ x+ x+x
    BOOST_MODE_MINUTES = "FH"
    CONTROL_SYSTEM_STATE = "MP"
    COOLING_MODE = "MK"
    FIREPLACE_MODE = "MB"
    FIREPLACE_MODE_MINUTES = "MC"
    MODE_FAN = "MF"
    MODE_HEATER = "MH"
    MODE_HEATER_POWER_RATING = "MG"
    MODE_TEMPERATURE = "MT"
    TARGET_TEMPERATURE_COOL = "TF"
    TARGET_TEMPERATURE_ECONOMY = "TE"
    TARGET_TEMPERATURE_NORMAL = "TD"

    NONE = ""
